#
# Licensed to Elasticsearch under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Elasticsearch licenses this file to you under
# the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# Get an Okta policy and its rules

import logging.config
from textwrap import dedent

import click

from dorothy.core import (
    get_policy_object,
    print_policy_object,
    print_module_info,
    set_module_options,
    reset_module_options,
    check_module_options,
    write_json_file,
    index_event,
)
from dorothy.modules.discovery.discovery import discovery

LOGGER = logging.getLogger(__name__)
MODULE_DESCRIPTION = "Get an Okta policy and its rules"
TACTICS = ["Discovery"]

MODULE_OPTIONS = {"id": {"value": None, "required": True, "help": "The unique ID for policy"}}


@discovery.subshell(name="get-policy")
def get_policy():
    """Get an Okta policy and its rules.

    This module gets a policy object and its rules using the policy's ID."""


@get_policy.command()
def info():
    """Show available options and their current values for this module"""

    print_module_info(MODULE_OPTIONS)


@get_policy.command()
@click.pass_context
@click.option("--id", help=MODULE_OPTIONS["id"]["help"])
def set(ctx, **kwargs):
    """Set one or more options for this module"""

    if all(value is None for value in kwargs.values()):
        return click.echo(ctx.get_help())

    else:
        global MODULE_OPTIONS
        MODULE_OPTIONS = set_module_options(MODULE_OPTIONS, kwargs)


@get_policy.command()
def reset():
    """Reset the options for this module"""

    global MODULE_OPTIONS
    MODULE_OPTIONS = reset_module_options(MODULE_OPTIONS)


@get_policy.command()
@click.pass_context
def execute(ctx):
    """Execute this module with the configured options"""

    error = check_module_options(MODULE_OPTIONS)

    if error:
        return

    msg = f'Attempting to get policy object for policy ID {MODULE_OPTIONS["id"]["value"]}'
    LOGGER.info(msg)
    index_event(ctx.obj.es, module=__name__, event_type="INFO", event=msg)
    click.echo(f"[*] {msg}")

    policy = get_policy_object(ctx, MODULE_OPTIONS["id"]["value"], rules=True)

    if policy:
        print_policy_object(policy)
        if click.confirm(f"[*] Do you want to save policy {policy['id']} ({policy['name']}) to a file?", default=True):
            file_path = f'{ctx.obj.data_dir}/{ctx.obj.profile_id}_policy_{policy["id"]}'
            write_json_file(file_path, policy)
